/*
 Copyright (c) 2015 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#ifndef GEOMETRY_HPP
#define GEOMETRY_HPP

#include <array>
#include <algorithm>
#include <numeric>
#include <cmath>
#include <functional>
#include <random>
#include <cassert>
#include <utility>

namespace geom {

constexpr long double PIl{3.1415926535897932384626433832795028841971693993751L};
constexpr double PI = double(PIl);
constexpr float PIf = float(PI);

// identity matrix
constexpr std::array<std::array<double, 3>, 3> identity3 = {
    {{{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}}};
constexpr std::array<std::array<float, 3>, 3> identity3f = {
    {{{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}}};
constexpr std::array<std::array<long double, 3>, 3> identity3l = {
    {{{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}}};

// vector subtraction, alias-safe
template <class T, size_t N>
inline void sub(const std::array<T, N>& a, const std::array<T, N>& b,
                std::array<T, N>& c) {
  std::transform(a.cbegin(), a.cend(), b.cbegin(), c.begin(), std::minus<T>());
}

// vector subtraction, alias-safe
template <class Inp1, class Inp2, class Out>
inline void sub(const Inp1 first, const Inp1 last, const Inp2 first2,
                Out dest) {
  std::transform(first, last, first2, dest,
                 std::minus<typename std::iterator_traits<Inp1>::value_type>());
}

// 3D vector subtraction
template <class T>
inline std::array<T, 3> sub(const std::array<T, 3>& a,
                            const std::array<T, 3>& b) {
  return {{a[0] - b[0], a[1] - b[1], a[2] - b[2]}};
}

// vector addition, alias-safe
template <class T, size_t N>
inline void add(const std::array<T, N>& a, const std::array<T, N>& b,
                std::array<T, N>& c) {
  std::transform(a.cbegin(), a.cend(), b.cbegin(), c.begin(), std::plus<T>());
}

// vector addition, alias-safe
template <class Inp1, class Inp2, class Out>
inline void add(const Inp1 first, const Inp1 last, const Inp2 first2,
                Out dest) {
  std::transform(first, last, first2, dest,
                 std::plus<typename std::iterator_traits<Inp1>::value_type>());
}

// 3D vector addition
template <class T>
inline std::array<T, 3> add(const std::array<T, 3>& a,
                            const std::array<T, 3>& b) {
  return {{a[0] + b[0], a[1] + b[1], a[2] + b[2]}};
}

// 3D vector cross product
template <class T>
inline std::array<T, 3> cross(const std::array<T, 3>& a,
                              const std::array<T, 3>& b) {
  return {{a[1] * b[2] - a[2] * b[1], a[2] * b[0] - a[0] * b[2],
           a[0] * b[1] - a[1] * b[0]}};
}

// length of vector
template <class T, size_t N>
inline T length(const std::array<T, N>& v) {
  return std::sqrt(std::inner_product(v.cbegin(), v.cend(), v.cbegin(), T(0)));
}

// length of vector
template <class T, size_t N>
inline T lengthsq(const std::array<T, N>& v) {
  return std::inner_product(v.cbegin(), v.cend(), v.cbegin(), T(0));
}

// multiply vector by scalar, alias-safe
template <class T, size_t N>
inline void scale(const std::array<T, N>& v1, T sc, std::array<T, N>& v2) {
  std::transform(v1.cbegin(), v1.cend(), begin(v2),
                 [sc](T val) { return val * sc; });
}

// dot product
template <class T, size_t N>
inline T dot(const std::array<T, N>& v1, const std::array<T, N>& v2) {
  return std::inner_product(v1.cbegin(), v1.cend(), v2.cbegin(), T(0));
}

// norm
template <class T, size_t N>
inline T norm(const std::array<T, N>& v) {
  return std::sqrt(std::inner_product(v.cbegin(), v.cend(), v.cbegin(), T(0)));
}

// normalize
template <class T, size_t N>
inline void normalize(const std::array<T, N>& v1, std::array<T, N>& v2) {
  T n = norm(v1);
  scale(v1, 1 / n, v2);
}

// square of distance between two points
template <class T, size_t N>
inline T distsq(const std::array<T, N>& a, const std::array<T, N>& b) {
  return std::inner_product(a.cbegin(), a.cend(), b.cbegin(), T(0),
                            std::plus<T>(),
                            [](T x, T y) { return (x - y) * (x - y); });
}

// distance between two N-dimensional points
template <class T, size_t N>
inline T dist(const std::array<T, N>& a, const std::array<T, N>& b) {
  return std::sqrt(distsq(a, b));
}

// clamp vector component values to a minimum value (in-place)
template <class T, size_t N>
inline void clamp_min(std::array<T, N>& a, T val) {
  std::replace_if(a.begin(), a.end(), [=](const T x) { return x < val; }, val);
}

// clamp to a maximum value (in-place)
template <class T, size_t N>
inline void clamp_max(std::array<T, N>& a, T val) {
  std::replace_if(a.begin(), a.end(), [=](const T x) { return x > val; }, val);
}

// apply fmod to all components of a vector (in-place)
template <class T, size_t N>
inline void fmod(std::array<T, N>& a, T val) {
  std::transform(a.cbegin(), a.cend(), a.begin(),
                 [=](const T x) { return std::fmod(x, val); });
}

// interpolation from a toward b
template <class T, size_t N>
inline void interpolate(const std::array<T, N>& a, const std::array<T, N>& b,
                        T frac, std::array<T, N>& c) {
  T frac2 = 1 - frac;
  std::transform(
      a.cbegin(), a.cend(), b.cbegin(), c.begin(),
      [frac, frac2](T val1, T val2) { return frac * val1 + frac2 * val2; });
}

// quaternion interpolation, alias-safe -- assumes normalized quaternions
template <class T>
void slerp(const std::array<T, 4>& q1, const std::array<T, 4>& q2, T alpha,
           std::array<T, 4>& q3) {

  assert(alpha >= T(0) && alpha <= T(1));

  // cosine (dot product)
  T cos_t = dot(q1, q2);

  // same quaternion (avoid domain error)
  if (std::abs(cos_t) >= 1.0) {
    q3 = q1;
    return;
  }
  T t = std::acos(cos_t);
  T sin_t = std::sin(t);

  // same quaternion (avoid zero-div)
  if (sin_t == 0.0) {
    q3 = q1;
    return;
  }

  T a = std::sin((1.0 - alpha) * t) / sin_t;
  T b = std::sin(alpha * t) / sin_t;
  // shortest path
  // if (cos_t < 0.0) b = -b;

  q3[0] = a * q1[0] + b * q2[0];
  q3[1] = a * q1[1] + b * q2[1];
  q3[2] = a * q1[2] + b * q2[2];
  q3[3] = a * q1[3] + b * q2[3];
}

// rotation matrix from quaternion
template <class T>
std::array<std::array<T, 3>, 3> make_matrix33(const std::array<T, 4>& q) {
  T x = q[0], y = q[1], z = q[2], w = q[3];
  T n = x * x + y * y + z * z + w * w;
  T s = (n > 0.0) ? (2.0 / n) : 0.0;

  T xs = x * s, ys = y * s, zs = z * s;
  T wx = w * xs, wy = w * ys, wz = w * zs;
  T xx = x * xs, xy = x * ys, xz = x * zs;
  T yy = y * ys, yz = y * zs, zz = z * zs;

  return {{{{1.0 - (yy + zz), xy - wz, xz + wy}},
           {{xy + wz, 1.0 - (xx + zz), yz - wx}},
           {{xz - wy, yz + wx, 1.0 - (xx + yy)}}}};
}

template <class T>
inline std::array<std::array<T, 4>, 4>
make_matrix44(const std::array<std::array<T, 3>, 3>& m) {
  return {{{{m[0][0], m[0][1], m[0][2], T(0)}},
           {{m[1][0], m[1][1], m[1][2], T(0)}},
           {{m[2][0], m[2][1], m[2][2], T(0)}},
           {{T(0), T(0), T(0), T(0)}}}};
}

template <class T>
inline std::array<std::array<T, 3>, 3>
make_matrix33(const std::array<std::array<T, 4>, 4>& m) {
  return {{{{m[0][0], m[0][1], m[0][2]}},
           {{m[1][0], m[1][1], m[1][2]}},
           {{m[2][0], m[2][1], m[2][2]}}}};
}

// create transformation from rotation, translation, scale
template <class T, class T1, class T2>
std::array<std::array<T, 4>, 4>
make_matrix44(const std::array<T, 4>& q, const std::array<T1, 3> t, T2 scale) {
  T x = q[0], y = q[1], z = q[2], w = q[3];
  T n = x * x + y * y + z * z + w * w;
  T s = (n > 0.0) ? (2.0 / n) : 0.0;

  T xs = x * s, ys = y * s, zs = z * s;
  T wx = w * xs, wy = w * ys, wz = w * zs;
  T xx = x * xs, xy = x * ys, xz = x * zs;
  T yy = y * ys, yz = y * zs, zz = z * zs;

  return {{{{scale * (1.0 - (yy + zz)), scale * (xy - wz), scale * (xz + wy),
             t[0]}},
           {{scale * (xy + wz), scale * (1.0 - (xx + zz)), scale * (yz - wx),
             t[1]}},
           {{scale * (xz - wy), scale * (yz + wx), scale * (1.0 - (xx + yy)),
             t[2]}},
           {{T(0), T(0), T(0), T(1.0)}}}};
}

// create transformation from rotation
template <class T>
std::array<std::array<T, 4>, 4> make_matrix44(const std::array<T, 4>& q) {
  T x = q[0], y = q[1], z = q[2], w = q[3];
  T n = x * x + y * y + z * z + w * w;
  T s = (n > 0.0) ? (2.0 / n) : 0.0;

  T xs = x * s, ys = y * s, zs = z * s;
  T wx = w * xs, wy = w * ys, wz = w * zs;
  T xx = x * xs, xy = x * ys, xz = x * zs;
  T yy = y * ys, yz = y * zs, zz = z * zs;

  return {{{{1.0 - (yy + zz), xy - wz, xz + wy, T(0)}},
           {{xy + wz, 1.0 - (xx + zz), yz - wx, T(0)}},
           {{xz - wy, yz + wx, 1.0 - (xx + yy), T(0)}},
           {{T(0), T(0), T(0), T(1.0)}}}};
}

// general matrix multiplication
// NOT alias-safe
template <class T1, size_t N1, size_t N, class T2, size_t N2, class T3>
void mul(const std::array<std::array<T1, N>, N1>& a,
         const std::array<std::array<T2, N2>, N>& b,
         std::array<std::array<T3, N1>, N2>& c) {
  for (auto i = size_t{0}; i < N1; ++i) {
    for (auto j = size_t{0}; j < N2; ++j) {
      T3 sum = 0;
      for (auto k = size_t{0}; k < N; ++k) {
        sum += static_cast<T3>(a[i][k] * b[k][j]);
      }
      c[i][j] = sum;
    }
  }
}

// general matrix multiplication
// NOT alias-safe
template <class T1, size_t N1, size_t N, class T2, size_t N2, class T3>
void mul_transpose_right(const std::array<std::array<T1, N>, N1>& a,
                         const std::array<std::array<T2, N>, N2>& b,
                         std::array<std::array<T3, N1>, N2>& c) {
  for (auto i = size_t{0}; i < N1; ++i) {
    for (auto j = size_t{0}; j < N2; ++j) {
      T3 sum = 0;
      for (auto k = size_t{0}; k < N; ++k) {
        sum += static_cast<T3>(a[i][k] * b[j][k]);
      }
      c[i][j] = sum;
    }
  }
}

// general matrix-vector multiplication
// NOT alias-safe
template <class T1, size_t N1, size_t N, class T2, class T3,
          class = typename std::enable_if<std::is_floating_point<T2>::value,
                                          T2>::type,
          class = typename std::enable_if<std::is_floating_point<T3>::value,
                                          T3>::type>
void mul(const std::array<std::array<T1, N>, N1>& a, const std::array<T2, N>& b,
         std::array<T3, N1>& c) {
  for (auto i = size_t{0}; i < N1; ++i) {
    T3 sum = 0;
    for (auto k = size_t{0}; k < N; ++k) {
      sum += static_cast<T3>(a[i][k] * b[k]);
    }
    c[i] = sum;
  }
}

// apply transformation matrix to a 3-D vector
template <class T>
std::array<T, 3> transform(const std::array<std::array<T, 4>, 4>& m,
                           const std::array<T, 3>& v) {
  return {{m[0][0] * v[0] + m[0][1] * v[1] + m[0][2] * v[2] + m[0][3],
           m[1][0] * v[0] + m[1][1] * v[1] + m[1][2] * v[2] + m[1][3],
           m[2][0] * v[0] + m[2][1] * v[1] + m[2][2] * v[2] + m[2][3]}};
}

// multiply matrix by scalar, in-place
template <class T, size_t N1, size_t N2, class T2>
void scale(std::array<std::array<T, N1>, N2>& m, T2 scale) {
  for (auto& r : m) {
    for (T& rc : r)
      rc *= static_cast<T>(scale);
  }
}

// transpose a square matrix in-place
template <class T, size_t N>
void transpose(std::array<std::array<T, N>, N>& m) {
  for (auto i = size_t{0}; i < (N - 1); ++i) {
    for (auto j = i + 1; j < N; ++j) {
      std::swap(m[i][j], m[j][i]);
    }
  }
}

template <class T>
T determinant(const std::array<std::array<T, 3>, 3>& m) {
  // less *,+,- calculation than expanded expression.
  return m[0][0] * (m[1][1] * m[2][2] - m[2][1] * m[1][2]) -
         m[0][1] * (m[1][0] * m[2][2] - m[2][0] * m[1][2]) +
         m[0][2] * (m[1][0] * m[2][1] - m[2][0] * m[1][1]);
}

// 3 x 3 matrix inversion, alias-safe
template <class T>
void invert(const std::array<std::array<T, 3>, 3>& m,
            std::array<std::array<T, 3>, 3>& dest) {
  // zero div could occur
  T s = 1.0 / determinant(m);
  dest = {{{{s * (m[1][1] * m[2][2] - m[1][2] * m[2][1]),
             s * (m[0][2] * m[2][1] - m[0][1] * m[2][2]),
             s * (m[0][1] * m[1][2] - m[0][2] * m[1][1])}},
           {{s * (m[1][2] * m[2][0] - m[1][0] * m[2][2]),
             s * (m[0][0] * m[2][2] - m[0][2] * m[2][0]),
             s * (m[0][2] * m[1][0] - m[0][0] * m[1][2])}},
           {{s * (m[1][0] * m[2][1] - m[1][1] * m[2][0]),
             s * (m[0][1] * m[2][0] - m[0][0] * m[2][1]),
             s * (m[0][0] * m[1][1] - m[0][1] * m[1][0])}}}};
}

// multiply two quaternions, alias-safe
template <class T>
void mulquat(const std::array<T, 4>& q1, const std::array<T, 4>& q2,
             std::array<T, 4>& dest) {
  dest = {{q1[0] * q2[3] + q1[3] * q2[0] + q1[1] * q2[2] - q1[2] * q2[1],
           q1[1] * q2[3] + q1[3] * q2[1] + q1[2] * q2[0] - q1[0] * q2[2],
           q1[2] * q2[3] + q1[3] * q2[2] + q1[0] * q2[1] - q1[1] * q2[0],
           q1[3] * q2[3] - q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2]}};
}

template <typename T>
std::array<T, 4> make_quat(const std::array<std::array<T, 3>, 3>& m) {
  // From Ken Shoemake
  // (ftp://ftp.cis.upenn.edu/pub/graphics/shoemake)

  std::array<T, 4> q;
  T s;
  T tr = m[0][0] + m[1][1] + m[2][2];
  if (tr >= 0.0) {
    s = std::sqrt(tr + 1.0);
    q[3] = s * 0.5;
    s = 0.5 / s;
    q[0] = (m[2][1] - m[1][2]) * s;
    q[1] = (m[0][2] - m[2][0]) * s;
    q[2] = (m[1][0] - m[0][1]) * s;
  } else {
    T maxm = std::max(m[0][0], std::max(m[1][1], m[2][2]));
    if (maxm == m[0][0]) {
      s = std::sqrt(m[0][0] - (m[1][1] + m[2][2]) + 1.0);
      q[0] = s * 0.5;
      s = 0.5 / s;
      q[1] = (m[0][1] + m[1][0]) * s;
      q[2] = (m[2][0] + m[0][2]) * s;
      q[3] = (m[2][1] - m[1][2]) * s;
    } else if (maxm == m[1][1]) {
      s = std::sqrt(m[1][1] - (m[2][2] + m[0][0]) + 1.0);
      q[1] = s * 0.5;
      s = 0.5 / s;
      q[2] = (m[1][2] + m[2][1]) * s;
      q[0] = (m[0][1] + m[1][0]) * s;
      q[3] = (m[0][2] - m[2][0]) * s;
    } else {
      s = std::sqrt(m[2][2] - (m[0][0] + m[1][1]) + 1.0);
      q[2] = s * 0.5;
      s = 0.5 / s;
      q[0] = (m[2][0] + m[0][2]) * s;
      q[1] = (m[1][2] + m[2][1]) * s;
      q[3] = (m[1][0] - m[0][1]) * s;
    }
  }
  return q;
}

template <typename T, typename RNG_T>
std::array<T, 4> make_quat(std::normal_distribution<T>& gauss, RNG_T& rng) {
  std::array<T, 4> quat = {{gauss(rng), gauss(rng), gauss(rng), gauss(rng)}};
  normalize(quat, quat);
  return quat;
}

// create a quaternion with a random rotation using
// three uniform random numbers in the range [0->1)
template <typename T, typename RNG_T>
std::array<T, 4> make_quat(std::uniform_real_distribution<T>& uniform,
                           RNG_T& rng) {
  T r1 = uniform(rng);
  T s1 = std::sqrt(1.0 - r1);
  T s2 = std::sqrt(r1);
  T p1 = 2.0 * PI * uniform(rng);
  T p2 = 2.0 * PI * uniform(rng);
  std::array<T, 4> quat;
  quat[0] = std::sin(p1) * s1;
  quat[1] = std::cos(p1) * s1;
  quat[2] = std::sin(p2) * s2;
  quat[3] = std::cos(p2) * s2;
  // normalize(quat, quat);
  return quat;
}
}

#endif
