/*
 Copyright (c) 2015 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#ifndef GRID3_HPP
#define GRID3_HPP

#include <array>
#include <cstdint>

template <class T>
class grid3 {
public:
  grid3(float xlow, float ylow, float zlow, float xincr, float yincr,
        float zincr, size_t xsize, size_t ysize, size_t zsize);
  virtual ~grid3();

  inline std::array<float, 3> origin() const noexcept;
  inline std::array<float, 3> spacing() const noexcept;
  inline std::array<size_t, 3> extents() const noexcept;

  // access the raw data through a pointer to the start of a one-dimensional
  // array
  // provided for interoperability with non-c++ code
  const T* data() const;

  // get the value at the nearest point in the grid
  // does unchecked access into the grid
  template <class RealT>
  T nearest_value(RealT x, RealT y, RealT z) const noexcept;

  template <class Vec3T>
  T nearest_value(const Vec3T& xyz) const noexcept;

  // get the interpolated value
  // does unchecked access into the grid
  template <class RealT>
  T linear_interpolated_value(RealT x, RealT y, RealT z) const noexcept;

  template <class Vec3T>
  T linear_interpolated_value(const Vec3T& xyz) const noexcept;

  // set the value at the nearest grid point
  // does unchecked access into the grid
  void set_value(double a, double b, double c, T val) noexcept;

private:
  // int_fast32_t is used for indexing rather than size_t because it can be
  // faster with some compilers
  using index_t = std::int_fast32_t;

  template <class RealT>
  static T lin_interp(const T& a, const T& b, RealT frac) noexcept;

  T* values;
  std::array<float, 3> orig;
  std::array<float, 3> incr;
  std::array<float, 3> inv_incr;
  index_t nx;
  index_t ny;
  index_t nz;
};

template <class T>
grid3<T>::grid3(float xlow, float ylow, float zlow, float xincr, float yincr,
                float zincr, size_t xsize, size_t ysize, size_t zsize) {
  values = new T[xsize * ysize * zsize];
  nx = index_t(xsize);
  ny = index_t(ysize);
  nz = index_t(zsize);
  orig = {{xlow, ylow, zlow}};
  incr = {{xincr, yincr, zincr}};
  inv_incr = {{1.0f / incr[0], 1.0f / incr[1], 1.0f / incr[2]}};
}

template <class T>
grid3<T>::~grid3() {
  delete[] values;
}

template <class T>
inline std::array<float, 3> grid3<T>::origin() const noexcept {
  return orig;
}

template <class T>
inline std::array<float, 3> grid3<T>::spacing() const noexcept {
  return incr;
}

template <class T>
inline std::array<size_t, 3> grid3<T>::extents() const noexcept {
  return {{nx, ny, nz}};
}

template <class T>
inline const T* grid3<T>::data() const {
  return values;
}

template <class T>
void grid3<T>::set_value(double x, double y, double z, T val) noexcept {
  auto xind = index_t((x - orig[0]) * inv_incr[0] + 0.5f);
  auto yind = index_t((y - orig[1]) * inv_incr[1] + 0.5f);
  auto zind = index_t((z - orig[2]) * inv_incr[2] + 0.5f);
  values[xind + nx * (yind + ny * zind)] = val;
}

template <class T>
template <class RealT>
T grid3<T>::nearest_value(RealT x, RealT y, RealT z) const noexcept {
  auto xind = index_t((x - orig[0]) * inv_incr[0] + 0.5f);
  auto yind = index_t((y - orig[1]) * inv_incr[1] + 0.5f);
  auto zind = index_t((z - orig[2]) * inv_incr[2] + 0.5f);
  return values[xind + nx * (yind + ny * zind)];
}

template <class T>
template <class Vec3T>
T grid3<T>::nearest_value(const Vec3T& xyz) const noexcept {
  auto xind = index_t((xyz[0] - orig[0]) * inv_incr[0] + 0.5f);
  auto yind = index_t((xyz[1] - orig[1]) * inv_incr[1] + 0.5f);
  auto zind = index_t((xyz[2] - orig[2]) * inv_incr[2] + 0.5f);
  return values[xind + nx * (yind + ny * zind)];
}

template <class T>
template <class RealT>
T grid3<T>::lin_interp(const T& a, const T& b, RealT frac) noexcept {
  // the less numerically stable form is used here
  // as it is significantly faster with some compilers
  // and the precision of the more stable equation isn't needed
  // return (1 - frac) * a + frac * b;
  return a + (b - a) * frac;
}

template <class T>
template <class RealT>
T grid3<T>::linear_interpolated_value(RealT x, RealT y, RealT z) const
    noexcept {
  RealT xfrac = (x - orig[0]) * inv_incr[0];
  RealT yfrac = (y - orig[1]) * inv_incr[1];
  RealT zfrac = (z - orig[2]) * inv_incr[2];
  auto xlow = index_t(xfrac);
  auto ylow = index_t(yfrac);
  auto zlow = index_t(zfrac);
  xfrac -= xlow;
  yfrac -= ylow;
  zfrac -= zlow;
  // factoring out common terms doesn't seem to speed this up
  T x_bot_val = lin_interp(values[xlow + nx * (ylow + ny * zlow)],
                           values[xlow + 1 + nx * (ylow + ny * zlow)], xfrac);
  T x_top_val =
      lin_interp(values[xlow + nx * (ylow + 1 + ny * zlow)],
                 values[xlow + 1 + nx * (ylow + 1 + ny * zlow)], xfrac);
  T xy_bot_val =
      lin_interp(values[xlow + nx * (ylow + ny * (zlow + 1))],
                 values[xlow + 1 + nx * (ylow + ny * (zlow + 1))], xfrac);
  T xy_top_val =
      lin_interp(values[xlow + nx * (ylow + 1 + ny * (zlow + 1))],
                 values[xlow + 1 + nx * (ylow + 1 + ny * (zlow + 1))], xfrac);
  T xval = lin_interp(x_bot_val, x_top_val, yfrac);
  T xyval = lin_interp(xy_bot_val, xy_top_val, yfrac);
  return lin_interp(xval, xyval, zfrac);
}

template <class T>
template <class Vec3T>
T grid3<T>::linear_interpolated_value(const Vec3T& xyz) const noexcept {
  typename Vec3T::value_type xfrac = (xyz[0] - orig[0]) * inv_incr[0];
  typename Vec3T::value_type yfrac = (xyz[1] - orig[1]) * inv_incr[1];
  typename Vec3T::value_type zfrac = (xyz[2] - orig[2]) * inv_incr[2];
  auto xlow = index_t(xfrac);
  auto ylow = index_t(yfrac);
  auto zlow = index_t(zfrac);
  xfrac -= xlow;
  yfrac -= ylow;
  zfrac -= zlow;
  // factoring out common terms doesn't speed this up with current compilers
  T x_bot_val = lin_interp(values[xlow + nx * (ylow + ny * zlow)],
                           values[xlow + 1 + nx * (ylow + ny * zlow)], xfrac);
  T x_top_val =
      lin_interp(values[xlow + nx * (ylow + 1 + ny * zlow)],
                 values[xlow + 1 + nx * (ylow + 1 + ny * zlow)], xfrac);
  T xy_bot_val =
      lin_interp(values[xlow + nx * (ylow + ny * (zlow + 1))],
                 values[xlow + 1 + nx * (ylow + ny * (zlow + 1))], xfrac);
  T xy_top_val =
      lin_interp(values[xlow + nx * (ylow + 1 + ny * (zlow + 1))],
                 values[xlow + 1 + nx * (ylow + 1 + ny * (zlow + 1))], xfrac);
  T xval = lin_interp(x_bot_val, x_top_val, yfrac);
  T xyval = lin_interp(xy_bot_val, xy_top_val, yfrac);
  return lin_interp(xval, xyval, zfrac);
}

#endif
