/*
 Copyright (c) 2015 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#include "epmr.hpp"

#if defined(_MSC_VER)
#include "getopt.h"
#else
#include <unistd.h>
#endif
#include <stdexcept>
#include <fstream>
#include <new>
#include <string>
#include <cstring>
#include <cstdlib>
#include <unordered_map>

using std::cout;
using std::cerr;
using std::endl;
using std::unordered_map;
using std::string;

unordered_map<string, string> set_options(int argc, char* argv[]);
void print_usage(std::ostream& output_stream);
void set_config_file_options(const char* optarg,
                             unordered_map<string, string>& options);
void check_integer(char* arg);
void check_numeric(char* arg);

int main(int argc, char* argv[]) {
  auto exitcode = EXIT_FAILURE;
  std::ios_base::sync_with_stdio(false);
  try {
    const auto options = set_options(argc, argv);
    epmr(options, cout);
    exitcode = EXIT_SUCCESS;
  } catch (const std::ios_base::failure& e) {
    cerr << argv[0] << ": An I/O error has occurred: " << e.what()
         << "\nExiting...\n";
  } catch (const std::invalid_argument& e) {
    cerr << argv[0] << ": Error parsing numeric value from input: " << e.what()
         << "\nExiting...\n";
  } catch (const std::bad_alloc&) {
    cerr << argv[0] << ": Out of memory.\nExiting...\n";
  } catch (const std::exception& e) {
    cerr << argv[0] << ": Error: " << e.what() << "\nExiting...\n";
  } catch (const char* str) {
    cerr << argv[0] << ": Error: " << str << "\nExiting...\n";
  } catch (...) {
    cerr << argv[0] << ": An error has occurred.\nExiting...\n";
  }
  return exitcode;
}

unordered_map<string, string> set_options(int argc, char* argv[]) {
  unordered_map<string, string> options;
  auto using_config_file = false;
  auto ch = -1;
  while ((ch = getopt(argc, argv,
                      "Aab:Cce:F:g:h:l:M:m:n:o:p:P:RSs:Tt:Vv:w:Z:")) != -1) {
    switch (ch) {
      case 'a':
        options["assembly"] = "true";
        options["optimize.assembly"] = "false";
        options["optimize.model"] = "false";
        break;
      case 'A':
        options["assembly"] = "true";
        options["optimize.assembly"] = "true";
        options["optimize.model"] = "false";
        break;
      case 'b':
        // options["bump.threshold"] = optarg;
        throw std::runtime_error(
            "option -b is not implemented in this version of the program.");
        break;
      case 'c':
        options["pointgroupsearch"] = "true";
        options["optimize.spacegroup"] = "false";
        break;
      case 'C':
        options["pointgroupsearch"] = "true";
        options["optimize.spacegroup"] = "true";
        break;
      case 'e':
        check_integer(optarg);
        options["random.seed"] = optarg;
        break;
      case 'F':
        set_config_file_options(optarg, options);
        using_config_file = true;
        break;
      case 'g':
        check_integer(optarg);
        options["evolution.generations"] = optarg;
        break;
      case 'h':
        check_numeric(optarg);
        options["resolution.high"] = optarg;
        break;
      case 'l':
        check_numeric(optarg);
        options["resolution.low"] = optarg;
        break;
      case 'm':
        check_integer(optarg);
        options["copies"] = optarg;
        options["optimize.copies"] = "false";
        break;
      case 'M':
        check_integer(optarg);
        options["copies"] = optarg;
        options["optimize.copies"] = "true";
        break;
      case 'n':
        check_integer(optarg);
        options["attempts"] = optarg;
        break;
      case 'o':
        options["output.prefix"] = optarg;
        break;
      case 'p':
        check_integer(optarg);
        options["evolution.population"] = optarg;
        break;
      case 'P': {
        // parse a development-only option taking the form: name=value
        auto arg = string(optarg);
        auto pos = arg.find('=');
        if (pos == string::npos || pos == arg.size() - 1) {
          throw std::runtime_error(
              "incorrect parameter definition on command line: " + arg);
        }
        // split optarg into key and value
        options[arg.substr(0, pos)] = arg.substr(pos + 1);
        break;
      }
      case 'R':
        options["optimize.orientation"] = "true";
        options["optimize.position"] = "false";
        break;
      case 's':
        if (options["structure.source"].length() > 0) {
          options["structure.source"] += " ";
        }
        options["structure.source"] = options["structure.source"] + optarg;
        break;
      case 'S':
        options["optimize.model"] = "true";
        options["assembly"] = "false";
        options["optimize.assembly"] = "false";
        break;
      case 't':
        // options["target.threshold"] = optarg;
        throw std::runtime_error(
            "option -t is not implemented in this version of the program.");
        break;
      case 'T':
        options["optimize.position"] = "true";
        options["optimize.orientation"] = "false";
        break;
      case 'v': {
        check_integer(optarg);
        auto index = std::stoi(optarg);
        switch (index) {
          case 0:
            options["report.header"] = "false";
            options["report.inputs"] = "false";
            options["report.options"] = "false";
            options["report.steps"] = "false";
            options["report.evolution"] = "none";
            options["report.local"] = "false";
            options["report.stats"] = "none";
            options["report.solution"] = "false";
            options["report.time.total"] = "false";
            break;
          case 1:
            options["report.header"] = "true";
            options["report.inputs"] = "true";
            options["report.options"] = "true";
            options["report.steps"] = "true";
            options["report.evolution"] = "progress";
            options["report.local"] = "true";
            options["report.stats"] = "none";
            options["report.solution"] = "true";
            options["report.time.total"] = "true";
            break;
          case 2:
            options["report.header"] = "true";
            options["report.inputs"] = "true";
            options["report.options"] = "true";
            options["report.steps"] = "true";
            options["report.evolution"] = "generations";
            options["report.local"] = "true";
            options["report.stats"] = "none";
            options["report.solution"] = "true";
            options["report.time.total"] = "true";
            break;
          default:
            throw std::runtime_error("Invalid value for option -v");
            break;
        }
        break;
      }
      case 'V':
        print_epmr_header(cout);
        exit(EXIT_SUCCESS);
        break;
      case 'w': {
        check_integer(optarg);
        string levels[] = {"none", "best", "all"};
        auto index = std::stoi(optarg);
        if (index < 0)
          index = 0;
        if (index > 2)
          index = 2;
        options["output.level"] = levels[index];
        break;
      }
      case 'Z':
        options["data.mtz-column"] = optarg;
        break;
      default:
        // getopt has already printed a message
        throw std::runtime_error("Unrecognized option");
    }
  }
  argc -= optind;
  argv += optind;

  if ((argc < 1 && !using_config_file) || argc > 3) {
    print_usage(cerr);
    exit(EXIT_FAILURE);
  }
  if (argc > 0) {
    auto path = string{argv[0]};
    options["cell.source"] = path;
    path = argc > 1 ? argv[1] : "-";
    options["model.source"] = path;
    path = argc > 2 ? argv[2] : argv[0];
    options["data.source"] = path;
  }
  return options;
}

void set_config_file_options(const char* optarg,
                             unordered_map<string, string>& options) {
  std::ifstream config_stream{optarg};
  if (config_stream.fail()) {
    string errmsg = "unable to open the file ";
    throw std::ios_base::failure(errmsg + optarg);
  }
  config_stream.exceptions(std::ios_base::badbit);
  auto optstr = string{};
  while (config_stream >> optstr) {
    auto pos = optstr.find('=');
    if (pos == string::npos || pos == optstr.length() - 1) {
      throw std::runtime_error(
          "Incorrect parameter definition in configuration file: " + optstr);
    }
    // split optarg into key and value
    options[optstr.substr(0, pos)] = optstr.substr(pos + 1);
  }
}

void print_usage(std::ostream& output_stream) {
  output_stream
      << "Usage: epmr [options] sca-file pdb-file\n"
      << "       epmr [options] sca-file @model-list-file\n"
      << "       epmr [options] mtz-file pdb-file\n"
      << "       epmr [options] mtz-file @model-list-file\n"
      << "       epmr [options] cell-file pdb-file reflection-file\n"
      << "       epmr [options] cell-file @model-list-file reflection-file\n"
      << "Options:\n"
      << "   -a               treat models as an assembly, search with each "
         "sequentially\n"
      << "   -A               treat models as an assembly, search with all "
         "simultaneously\n" <<
      //    "   -b number        minimum packing distance without a bump penalty
      //    (" <<
      //    epmr_defaults.at("bump.threshold") << ")\n" <<
      "   -c               search sequentially over all related space groups\n"
      << "   -C               search all related space groups simultaneously\n"
      << "   -e number        seed for random number generator, "
         "0=system-supplied (" << epmr_defaults.at("random.seed") << ")\n"
      << "   -g number        number of generations ("
      << epmr_defaults.at("evolution.generations") << ")\n"
      << "   -h number        high resolution limit ("
      << epmr_defaults.at("resolution.high") << ")\n"
      << "   -l number        low resolution limit ("
      << epmr_defaults.at("resolution.low") << ")\n"
      << "   -m number        number of copies of model to find sequentially ("
      << epmr_defaults.at("copies") << ")\n"
      << "   -M number        number of copies of model to find simultaneously "
         "(" << epmr_defaults.at("copies") << ")\n"
      << "   -n number        number of search attempts ("
      << epmr_defaults.at("attempts") << ")\n"
      << "   -o filename      prefix for output coordinate files ("
      << epmr_defaults.at("output.prefix") << ")\n"
      << "   -p number        population size ("
      << epmr_defaults.at("evolution.population") << ")\n"
      << "   -R               rotation search only\n"
      << "   -s filename      file containing static structure\n"
      << "   -S               use all search models in competition with each "
         "other\n"
      //    "   -t number        threshold score to stop search (" <<
      //    epmr_defaults.at("target.threshold") << ")\n" <<
      << "   -T               translation search only\n"
      << "   -v number        verbosity: 0=none, 1=normal, 2=high (1)\n"
      << "   -V               print version information and exit\n"
      << "   -w number        write solution coordinates: 0=none, 1=best, "
         "2=all (" << epmr_defaults.at("output.level") << ")\n"
      << "   -Z label         MTZ data column (structure factors or "
         "intensities)" << endl;
}

void check_integer(char* arg) {
  char* p;
  std::strtol(arg, &p, 10);
  if (*p != '\0') {
    auto msg = string{"command line parameter is not an integer: "};
    throw std::runtime_error(msg + arg);
  }
}

void check_numeric(char* arg) {
  char* p;
  std::strtod(arg, &p);
  if (*p != '\0') {
    auto msg = string{"command line parameter is not a number: "};
    throw std::runtime_error(msg + arg);
  }
}
