
/*
 Copyright (c) 2015 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#ifndef PLATFORM_HPP
#define PLATFORM_HPP

#include "evolutionaryoptimizer.hpp"
#include "solution.hpp"
#include <vector>
#include <future>
#include <algorithm>

enum class parallelism { None, Async, GCD, PPL, OpenMP, AMP, OpenCL };

#if defined(__APPLE__) && defined(__BLOCKS__)
const parallelism global_population_parallelism =
    parallelism::Async; // clang linker error when set to GCD and optimizing
const parallelism global_reflection_parallelism = parallelism::None;
const parallelism local_reflection_parallelism = parallelism::GCD;
#elif defined(_MSC_VER)
const parallelism global_population_parallelism = parallelism::PPL;
const parallelism global_reflection_parallelism = parallelism::None;
const parallelism local_reflection_parallelism = parallelism::PPL;
#elif defined(_OPENMP)
const parallelism global_population_parallelism = parallelism::OpenMP;
const parallelism global_reflection_parallelism = parallelism::None;
const parallelism local_reflection_parallelism = parallelism::OpenMP;
#else
const parallelism global_population_parallelism = parallelism::Async;
const parallelism global_reflection_parallelism = parallelism::None;
const parallelism local_reflection_parallelism = parallelism::Async;
#endif

template <parallelism ParallelType>
class eo_platform_config {
public:
  static void configure(optimization::evolutionary_optimizer<solution>& eo);
};

template <>
class eo_platform_config<parallelism::None> {
public:
  static void configure(optimization::evolutionary_optimizer<solution>& eo) {
    eo.score_population = [&eo](
        const std::vector<solution>& population, std::vector<double>& scores,
        const std::vector<size_t>& indices, size_t nsurv) {
      std::for_each(indices.cbegin() + nsurv, indices.cend(),
                    [&](size_t i) { scores[i] = eo.score(population[i]); });
    };
  }
};

template <>
class eo_platform_config<parallelism::Async> {
public:
  static void configure(optimization::evolutionary_optimizer<solution>& eo) {
    const auto nthr = std::max(1u, std::thread::hardware_concurrency());
    eo.score_population = [&eo, nthr](
        const std::vector<solution>& population, std::vector<double>& scores,
        const std::vector<size_t>& indices, size_t nsurv) {
      auto first = nsurv;
      auto last = population.size();
      auto stride = std::max(size_t{1}, (last - first) / nthr);
      auto futures = std::vector<std::future<void>>();
      while (first < last) {
        const size_t limit = std::min(last - first, stride) + first;
        futures.emplace_back(
            std::async(std::launch::async, [&, first, limit]() {
              for (auto i = first; i < limit; ++i) {
                scores[indices[i]] = eo.score(population[indices[i]]);
              }
            }));
        first = limit;
      }
      for (auto& fut : futures) {
        fut.wait();
      }
    };
  }
};

#if defined(__APPLE__) && defined(__BLOCKS__)
#include <dispatch/dispatch.h>

template <>
class eo_platform_config<parallelism::GCD> {
public:
  static void configure(optimization::evolutionary_optimizer<solution>& eo) {
    eo.score_population = [&eo](
        const std::vector<solution>& population, std::vector<double>& scores,
        const std::vector<size_t>& indices, size_t nsurv) {
      static auto queue = dispatch_get_global_queue(0, 0);
      dispatch_apply(indices.size() - nsurv, queue, ^(size_t i) {
        scores[indices[i + nsurv]] = eo.score(population[indices[i + nsurv]]);
      });
    };
  }
};
#endif

template <>
class eo_platform_config<parallelism::OpenMP> {
public:
  static void configure(optimization::evolutionary_optimizer<solution>& eo) {
    eo.score_population = [&eo](
        const std::vector<solution>& population, std::vector<double>& scores,
        const std::vector<size_t>& indices, size_t nsurv) {
      auto limit = int(population.size());
#pragma omp parallel for
      // some OMP implementations require int index
      for (auto i = int(nsurv); i < limit; ++i) {
        scores[indices[i]] = eo.score(population[indices[i]]);
      }
    };
  }
};

#ifdef _MSC_VER
#include <ppl.h>

template <>
class eo_platform_config<parallelism::PPL> {
public:
  static void configure(optimization::evolutionary_optimizer<solution>& eo) {
    eo.score_population = [&eo](
        const std::vector<solution>& population, std::vector<double>& scores,
        const std::vector<size_t>& indices, size_t nsurv) {
      auto limit = population.size();
      Concurrency::parallel_for(nsurv, limit, [&](size_t i) {
        scores[indices[i]] = eo.score(population[indices[i]]);
      });
    };
  }
};

#endif

#endif
